@@ -1,5 +1,6 @@
-# perturbations/base.py
-
+import numpy as np
+import itertools
+import copy
 from typing import Tuple, Dict, List, Optional, Callable, Sequence, Union, Iterable
 from nlpaug.base_augmenter import Augmenter
 
@@ -12,15 +13,51 @@ from text_explainability.utils import default_tokenizer, default_detokenizer
 
 
 def oneway_dictionary_mapping(instance: TextInstance,
-                              dictionary: Dict[str, str],
+                              dictionary: Dict[str, List[str]],
                               label_from: LT,
                               label_to: LT,
+                              n: int,
                               tokenizer,
                               detokenizer) -> Optional[Tuple[str, LT, LT]]:
     tokenized = tokenizer(instance.data)
-    new_tokenized = [dictionary[a] if a in dictionary.keys() else a for a in tokenized]
-    if tokenized != new_tokenized:
-        return detokenizer(new_tokenized), label_from, label_to
+
+    # Get all options
+    options = {i: dictionary[a] for i, a in enumerate(tokenized) if a in dictionary.keys()}
+    all_options = list(itertools.product(*options.values()))
+
+    # Pick up to N random replacements and apply them
+    for idx in set(np.random.randint(len(all_options), size=n)):
+        current_option = all_options[idx]
+        new_tokenized = copy.deepcopy(tokenized)
+        for i, option in enumerate(current_option):
+            new_tokenized[options.keys()[i]] = option
+        if tokenized != new_tokenized:
+            yield detokenizer(new_tokenized), label_from, label_to
+
+
+def one_to_one_dictionary_mapping(instance: TextInstance,
+                                  dictionary: Dict[str, List[str]],
+                                  label_from: LT,
+                                  label_to: LT,
+                                  tokenizer,
+                                  detokenizer):
+    return list(oneway_dictionary_mapping(instance, dictionary, label_from=label_from, label_to=label_to,
+                                          tokenizer=tokenizer, detokenizer=detokenizer, n=1))[0]
+
+
+def one_to_many_dictionary_mapping(instance: TextInstance,
+                                   dictionary: Dict[str, List[str]],
+                                   label_from: LT,
+                                   label_to: LT,
+                                   n: int,
+                                   tokenizer,
+                                   detokenizer):
+    res = list(oneway_dictionary_mapping(instance, dictionary, label_from=label_from, label_to=label_to,
+                                         tokenizer=tokenizer, detokenizer=detokenizer, n=n))
+    res = list(filter(None, res))
+    if len(res) == 0:
+        return None
+    return res
 
 
 def as_list(x):
@@ -43,6 +80,15 @@ class Perturbation(Readable):
         """
         self.perturbation_function = perturbation_function
 
+    @classmethod
+    def from_dictionary(cls, *args, **kwargs):
+        raise NotImplementedError('Implemented in subclasses.')  
+
+    @classmethod
+    def from_dict(cls, *args, **kwargs):
+        """Alias for `Perturbation.from_dictionary()`."""
+        return cls.from_dictionary(*args, **kwargs)
+
     @classmethod
     def from_function(cls,
                       function: Callable[[str], Optional[Union[str, Sequence[str]]]],
@@ -68,6 +114,15 @@ class Perturbation(Readable):
 
         return cls(perturbation_function)
 
+    @classmethod
+    def from_string(cls, *args, **kwargs):
+        raise NotImplementedError('Implemented in subclasses.')
+
+    @classmethod
+    def from_str(cls, *args, **kwargs):
+        """Alias for `Perturbation.from_string()`."""
+        return cls.from_string(*args, **kwargs)
+
     @text_instance
     def perturb(self, instance: TextInstance):
         raise NotImplementedError('Implemented in subclasses.')
@@ -128,20 +183,15 @@ class OneToOnePerturbation(Perturbation):
         # TODO: add case-sensitivity
         @text_instance
         def perturbation_function(instance: TextInstance) -> Optional[Tuple[str, LT, LT]]:
-            return oneway_dictionary_mapping(instance,
-                                             dictionary=dictionary,
-                                             label_from=label_from,
-                                             label_to=label_to,
-                                             tokenizer=tokenizer,
-                                             detokenizer=detokenizer)
+            return one_to_one_dictionary_mapping(instance,
+                                                 dictionary={k: as_list(v) for k, v in dictionary},
+                                                 label_from=label_from,
+                                                 label_to=label_to,
+                                                 tokenizer=tokenizer,
+                                                 detokenizer=detokenizer)
 
         return cls(perturbation_function)
 
-    @classmethod
-    def from_dict(cls, *args, **kwargs):
-        """Alias for `OneToOnePerturbation.from_dictionary()`."""
-        return cls.from_dictionary(*args, **kwargs)
-
     @classmethod
     def from_tuples(cls,
                     tuples: List[Tuple[str, str]],
@@ -174,28 +224,64 @@ class OneToOnePerturbation(Perturbation):
             detokenizer (Callable, optional): Function to detokenize tokens into instance data.
                 Defaults to default_detokenizer.        
         """
-        dictionary_from = dict(tuples)
-        dictionary_to = {v: k for k, v in tuples}
+        dictionary_from = {k: as_list(v) for k, v in tuples}
+        dictionary_to = {v: as_list(k) for k, v in tuples}
 
         @text_instance
         def perturbation_function(instance: TextInstance) -> Optional[Tuple[str, LT, LT]]:
-            first_res = oneway_dictionary_mapping(instance,
-                                                  dictionary=dictionary_from,
-                                                  label_from=label_from,
-                                                  label_to=label_to,
-                                                  tokenizer=tokenizer,
-                                                  detokenizer=detokenizer)
+            first_res = one_to_one_dictionary_mapping(instance,
+                                                      dictionary=dictionary_from,
+                                                      label_from=label_from,
+                                                      label_to=label_to,
+                                                      tokenizer=tokenizer,
+                                                      detokenizer=detokenizer)
             if first_res is not None:
                 return first_res
-            return oneway_dictionary_mapping(instance,
-                                             dictionary=dictionary_to,
-                                             label_from=label_to,
-                                             label_to=label_from,
-                                             tokenizer=tokenizer,
-                                             detokenizer=detokenizer)
+            return one_to_one_dictionary_mapping(instance,
+                                                 dictionary=dictionary_to,
+                                                 label_from=label_to,
+                                                 label_to=label_from,
+                                                 tokenizer=tokenizer,
+                                                 detokenizer=detokenizer)
 
         return cls(perturbation_function)
 
+    @classmethod
+    def from_list(cls,
+                  mapping_list: List[str],
+                  label_from: LT = 'original',
+                  label_to: LT = 'perturbed',
+                  tokenizer: Callable = default_tokenizer,
+                  detokenizer: Callable = default_detokenizer):
+        """Construct a `OneToOnePerturbation` from a list.
+
+        A function is constructed that aims to map any value in the list to any other value 
+        in the list.
+
+        Example:
+            For example, if list `['Amsterdam', 'Rotterdam', 'Utrecht']` is provided it aims to map 
+            'Amsterdam' to 'Rotterdam' or 'Utrecht', 'Rotterdam' to 'Amsterdam' to 'Utrecht' and 
+            'Utrecht' to 'Rotterdam' or 'Amsterdam'. If None of these is possible, it returns None.
+
+            >>> map_list = ['Amsterdam', 'Rotterdam', 'Utrecht']
+            >>> OneToOnePerturbation.from_list(map_list)
+
+        Args:
+            mapping_list (List[str]): Lookup list of tokens (e.g. words, characters).
+            label_from (LT): Attribute label of original instance (non-replaced).
+            label_to (LT): Attribute label of perturbed instance (replaced).
+            tokenizer (Callable, optional): Function to tokenize instance data (e.g. words, characters).
+                Defaults to default_tokenizer.
+            detokenizer (Callable, optional): Function to detokenize tokens into instance data.
+                Defaults to default_detokenizer.        
+        """
+        mapping_list = [(a, b) for a, b in set(list(itertools.combinations(mapping_list, 2)))]
+        return OneToManyPerturbation.from_tuples(mapping_list,
+                                                 label_from=label_from,
+                                                 label_to=label_to,
+                                                 tokenizer=tokenizer,
+                                                 detokenizer=detokenizer)
+
     @classmethod
     def from_string(cls,
                     prefix: Optional[str] = None,
@@ -264,11 +350,6 @@ class OneToOnePerturbation(Perturbation):
 
         return cls(perturbation_function)
 
-    @classmethod
-    def from_str(cls, *args, **kwargs):
-        """Alias for `OneToOnePerturbation.from_string()`."""
-        return cls.from_string(*args, **kwargs)
-
     @classmethod
     def from_nlpaug(cls,
                     augmenter: Augmenter,
@@ -377,10 +458,56 @@ class OneToManyPerturbation(Perturbation):
         import inspect
 
         if 'n' in inspect.signature(function).parameters:
-            from functools import partial
-            return super().from_function(partial(function, n=n))
+            return super().from_function(lambda x: function(x, n=n), label_from=label_from, label_to=label_to)
 
-        return super().from_function(lambda x: [function(x) for _ in range(n)])
+        def perform_n_times(instance):
+            perturbed = list(filter(None, [function(instance) for _ in range(n)]))
+            return None if len(perturbed) == 0 else perturbed
+
+        return super().from_function(lambda x: perform_n_times(x), label_from=label_from, label_to=label_to)
+
+    @classmethod
+    def from_dictionary(cls,
+                        dictionary: Dict[str, List[str]],
+                        label_from: LT,
+                        label_to: LT,
+                        n: int = 10,
+                        tokenizer: Callable = default_tokenizer,
+                        detokenizer: Callable = default_detokenizer):
+        """Construct a `OneToManyPerturbation` from a dictionary.
+
+        Example:
+            Replace the word 'good' (positive) with 'bad', 'mediocre', 'terrible' (negative) up to
+            5 times in each instance. The default tokenizer/detokenizer assumes word-level tokens:
+
+            >>> replacements = {'good': ['bad', 'mediocre', 'terrible']}
+            >>> OneToManyPerturbation.from_dictionary(replacement,
+            >>>                                       n=5,
+            >>>                                       label_from='positive',
+            >>>                                       label_to='negative')
+
+        Args:
+            dictionary (Dict[str, List[str]]): Lookup dictionary to map tokens (e.g. words, characters).
+            label_from (LT): Attribute label of original instance (left-hand side of dictionary).
+            label_to (LT): Attribute label of perturbed instance (right-hand side of dictionary).
+            n (int, optional): Number of instances to generate. Defaults to 10.
+            tokenizer (Callable, optional): Function to tokenize instance data (e.g. words, characters).
+                Defaults to default_tokenizer.
+            detokenizer (Callable, optional): Function to detokenize tokens into instance data.
+                Defaults to default_detokenizer.
+        """
+        # TODO: add case-sensitivity
+        @text_instance
+        def perturbation_function(instance: TextInstance) -> Optional[Tuple[str, LT, LT]]:
+            return one_to_many_dictionary_mapping(instance,
+                                                  dictionary={k: as_list(v) for k, v in dictionary},
+                                                  label_from=label_from,
+                                                  label_to=label_to,
+                                                  n=n,
+                                                  tokenizer=tokenizer,
+                                                  detokenizer=detokenizer)
+
+        return cls(perturbation_function)
 
     @classmethod
     def from_nlpaug(cls,
@@ -417,7 +544,7 @@ class OneToManyPerturbation(Perturbation):
             except:  # noqa: E722
                 return None
 
-        return cls.from_function(perturbation_function, label_from=label_from, label_to=label_to)
+        return cls.from_function(perturbation_function, n=n, label_from=label_from, label_to=label_to)
 
     @text_instance
     def perturb(self, instance: TextInstance) -> Optional[Sequence[Tuple[TextInstance, Sequence[Tuple[KT, LT]]]]]:
